# TODO: This file is a modified copy of the old gui/Platform.py
"""
Copyright 2008, 2009 Free Software Foundation, Inc.
This file is part of GNU Radio

SPDX-License-Identifier: GPL-2.0-or-later

"""


import sys
import os
from collections import ChainMap

from .Config import Config
from .components.canvas.flowgraph import Flowgraph
from .components.canvas.block import Block
from .components.canvas.port import Port
from .components.canvas.connection import Connection
from ..core.platform import Platform as CorePlatform


class Platform(CorePlatform):

    def __init__(self, *args, **kwargs):
        CorePlatform.__init__(self, *args, **kwargs)

        # Ensure conf directories
        gui_prefs_file = self.config.gui_prefs_file
        if not os.path.exists(os.path.dirname(gui_prefs_file)):
            os.mkdir(os.path.dirname(gui_prefs_file))

        self._move_old_pref_file()

    def get_prefs_file(self):
        return self.config.gui_prefs_file

    def _move_old_pref_file(self):
        gui_prefs_file = self.config.gui_prefs_file
        old_gui_prefs_file = os.environ.get(
            'GRC_PREFS_PATH', os.path.expanduser('~/.grc'))
        if gui_prefs_file == old_gui_prefs_file:
            return  # prefs file overridden with env var
        if os.path.exists(old_gui_prefs_file) and not os.path.exists(gui_prefs_file):
            try:
                import shutil
                shutil.move(old_gui_prefs_file, gui_prefs_file)
            except Exception as e:
                print(e, file=sys.stderr)

    ##############################################
    # Factories
    ##############################################
    Config = Config
    FlowGraph = Flowgraph
    Connection = Connection

    def new_block_class(self, **data):
        cls = CorePlatform.new_block_class(self, **data)
        return Block.make_cls_with_base(cls)

    block_classes_build_in = {key: Block.make_cls_with_base(cls)
                              for key, cls in CorePlatform.block_classes_build_in.items()}
    block_classes = ChainMap({}, block_classes_build_in)

    port_classes = {key: Port.make_cls_with_base(cls)
                    for key, cls in CorePlatform.port_classes.items()}
    # param_classes = {key: Param.make_cls_with_base(cls)
    #                 for key, cls in CorePlatform.param_classes.items()}



from __future__ import absolute_import, print_function

# Standard modules
import logging
import os
import yaml

from qtpy import QtCore, QtWidgets, uic

from .. import base
from ..properties import Paths


# Logging
log = logging.getLogger(f"grc.application.{__name__}")


class OOTBrowser(QtWidgets.QDialog, base.Component):
    data_role = QtCore.Qt.UserRole

    def __init__(self):
        super().__init__()
        uic.loadUi(Paths.RESOURCES + "/oot_browser.ui", self)

        self.setMinimumSize(600, 450)
        self.setModal(True)

        self.setWindowTitle("GRC OOT Modules")

        self.left_list.currentItemChanged.connect(self.populate_right_view)

        self.manifest_dir = os.path.join(Paths.RESOURCES, "manifests")

        if not os.path.exists(self.manifest_dir):
            return

        for f in os.listdir(self.manifest_dir):
            with open(os.path.join(self.manifest_dir, f), 'r', encoding='utf8') as manifest:
                text = manifest.read()
                yml, desc = text.split("---")
                data = yaml.safe_load(yml)
                data["description"] = desc
                self.validate(data)
                item = QtWidgets.QListWidgetItem()
                item.setText(data["title"])
                item.setData(self.data_role, data)

                self.left_list.addItem(item)

        self.left_list.setCurrentRow(0)

    def validate(self, module) -> bool:
        type_dict = {
            'title': str,
            'brief': str,
            'website': str,
            'dependencies': list,
            'repo': str,
            'copyright_owner': list,
            'gr_supported_version': list,
            'tags': list,
            'license': str,
            'description': str,
            'author': list
        }

        valid = True

        for key, val in type_dict.items():
            if key in module:
                if not type(module.get(key)) == val:
                    log.error(
                        f"OOT module {module.get('title')} has field {key}, but it's not the correct type. Expected {val}, got {type(module.get(key))}. Ignoring")
                    valid = False
            else:
                log.error(f"OOT module {module.get('title')} is missing field {key}. Ignoring")
                valid = False

        return valid

    def populate_right_view(self):
        module = self.left_list.currentItem().data(self.data_role)

        self.title_label.setText(f"{module.get('title')}")
        self.version_label.setText(f"<b>Version:</b> {module.get('version')}")
        self.brief_label.setText(module.get("brief"))
        if module.get('website'):
            self.website_label.setText(f"**Website:** {module.get('website')}")
        else:
            self.website_label.setText(f'**Website:** None')
        if module.get("dependencies"):
            self.dep_label.setText(f"<b>Dependencies:</b> {'; '.join(module.get('dependencies'))}")
        else:
            self.dep_label.setText("<b>Dependencies:</b> None")
        if module.get('repo'):
            self.repo_label.setText(f"**Repository:** {module.get('repo')}")
        else:
            self.repo_label.setText(f'**Repository:** None')
        if module.get("copyright_owner"):
            self.copyright_label.setText(f"<b>Copyright Owner:</b> {', '.join(module.get('copyright_owner'))}")
        else:
            self.copyright_label.setText("<b>Copyright Owner:</b> None")
        if type(module.get('gr_supported_version')) == list:
            self.supp_ver_label.setText(
                f"<b>Supported GNU Radio Versions:</b> {', '.join(module.get('gr_supported_version'))}")
        else:
            self.supp_ver_label.setText("<b>Supported GNU Radio Versions:</b> N/A")
            log.error(f"module {module.get('title')} has invalid manifest field gr_supported_version")

        self.tags_label.setText(f"<b>Tags:</b> {'; '.join(module.get('tags'))}")
        self.license_label.setText(f"<b>License:</b> {module.get('license')}")
        self.desc_label.setMarkdown("\n" + module.get("description").replace("\t", ""))
        self.author_label.setText(f"**Author(s):** {', '.join(module.get('author'))}")



# Copyright 2016 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# SPDX-License-Identifier: GPL-2.0-or-later
#

from ._lazy import lazy_property, nop_write

from .evaluated import (
    Evaluated,
    EvaluatedEnum,
    EvaluatedPInt,
    EvaluatedFlag,
    setup_names,
)




from .block_library import BlockLibrary
from .example_browser import ExampleBrowser, Worker
from .wiki_tab import WikiTab
from .console import Console
from .flowgraph_view import FlowgraphView
from .undoable_actions import ChangeStateAction
from .variable_editor import VariableEditor

# Import last since there are dependencies
from .window import MainWindow



# Copyright 2014-2020 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import, print_function

# Standard modules
import html
import logging
import textwrap

# Third-party  modules

from qtpy import QtCore, QtGui, QtWidgets

# Custom modules
from .. import base

# Shortcuts
Action = QtWidgets.QAction
Menu = QtWidgets.QMenu
Toolbar = QtWidgets.QToolBar
Icons = QtGui.QIcon.fromTheme
Keys = QtGui.QKeySequence

# Logging
log = logging.getLogger(f"grc.application.{__name__}")


HTML = '''
<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\"
\"http://www.w3.org/TR/REC-html40/strict.dtd\">
<html>
<head>
    <meta name=\"qrichtext\" content=\"1\" />
    <style type=\"text/css\">
        p, li {
            white-space: pre-wrap;
        }

        /* Not currently used because entries are wrapped with <pre> */
        body {
            /* font-family:  \'Ubuntu\'; */
            font-family:  \'MS Shell Dlg 2\';
            font-size:    10pt;
            font-weight:  400;
            font-style:   normal;
            font-color:   black;
            margin:       0px;
            text-indent:  0px;
            -qt-block-indent:   0;
        }

        pre {
            display: block;
            font-family: monospace;
            white-space: pre;
            font-color:  black;
            margin:      1em 0;
        }

    </style>
</head>
</html>
'''


class Console(QtWidgets.QDockWidget, base.Component):
    def __init__(self, level):
        super(Console, self).__init__()

        self.qsettings = self.app.qsettings

        self.setObjectName('console')
        self.setWindowTitle('Console')
        self.level = level

        # GUI Widgets

        # Create the layout widget
        container = QtWidgets.QWidget(self)
        container.setObjectName('console::container')
        self._container = container

        layout = QtWidgets.QHBoxLayout(container)
        layout.setObjectName('console::layout')
        layout.setSpacing(0)
        layout.setContentsMargins(5, 0, 5, 5)
        self._layout = layout

        # Console output widget
        text = QtWidgets.QTextEdit(container)
        text.setObjectName('console::text')
        text.setUndoRedoEnabled(False)
        text.setReadOnly(True)
        text.setCursorWidth(0)
        text.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard | QtCore.Qt.TextSelectableByMouse)
        text.setHtml(textwrap.dedent(HTML))
        self._text = text

        # Add widgets to the component
        layout.addWidget(text)
        container.setLayout(layout)
        self.setWidget(container)

        # Translation support

        # self.setWindowTitle(_translate("", "Library", None))
        # library.headerItem().setText(0, _translate("", "Blocks", None))
        # QtCore.QMetaObject.connectSlotsByName(blockLibraryDock)

        # Setup actions

        # TODO: Move to the base controller and set actions as class attributes
        # Automatically create the actions, menus and toolbars.
        # Child controllers need to call the register functions to integrate into the mainwindow
        self.actions = {}
        self.menus = {}
        self.toolbars = {}
        self.createActions(self.actions)
        self.createMenus(self.actions, self.menus)
        self.createToolbars(self.actions, self.toolbars)
        self.connectSlots()

        # Register the dock widget through the AppController.
        # The AppController then tries to find a saved dock location from the preferences
        # before calling the MainWindow Controller to add the widget.
        self.app.registerDockWidget(self, location=self.settings.window.CONSOLE_DOCK_LOCATION)

        # Register the menus
        self.app.registerMenu(self.menus["console"])

        # Register a new handler for the root logger that outputs messages of
        #  INFO and HIGHER to the reports view
        handler = ReportsHandler(self.add_line)
        handler.setLevel(self.level)

        # Need to add this handler to the parent of the controller's logger
        log.parent.addHandler(handler)
        self.handler = handler

        self.actions['show_level'].setChecked = True
        self.handler.show_level = True
        self.enabled = False

        if not self.qsettings.value("appearance/display_console", True, type=bool):
            self.hide()

    def enable(self):
        self.enabled = True

    # Actions

    def createActions(self, actions):
        ''' Defines all actions for this view. '''

        log.debug("Creating actions")

        # File Actions
        actions['save'] = Action(Icons("document-save"), _("save"), self, statusTip=_("save-tooltip"))
        actions['clear'] = Action(Icons("document-close"), _("clear"), self, statusTip=_("clear-tooltip"))
        actions['show_level'] = Action(_("show-level"), self, statusTip=_("show-level"), checkable=True, checked=True)

        actions['auto_scroll'] = Action(
            _("auto-scroll"), self, statusTip=_("auto-scroll"), checkable=True, checked=True)

    def createMenus(self, actions, menus):
        ''' Setup the view's menus '''

        log.debug("Creating menus")

        console_menu = QtWidgets.QMenu("&Console")
        console_menu.setObjectName("console::menu")

        # Not needed, we have FileHandler logging in main.py
        # console_menu.addAction(actions["save"])

        console_menu.addAction(actions["clear"])
        console_menu.addAction(actions["show_level"])
        console_menu.addAction(actions["auto_scroll"])
        menus["console"] = console_menu

    def createToolbars(self, actions, toolbars):
        log.debug("Creating toolbars")

    def add_line(self, line):
        # TODO: Support multiple columns for the HTML. DO better with the spacing
        #  and margins in the output
        if self.enabled:
            self._text.append(line)
            if self.actions["auto_scroll"].isChecked():
                self._text.verticalScrollBar().setValue(
                    self._text.verticalScrollBar().maximum())

    # Handlers for the view actions
    def clear_triggered(self):
        self._text.clear()

    def save_triggered(self):
        log.warning("Save reports not implemented")

    def show_level_toggled(self, checked):
        self.handler.show_level = checked


class ReportsHandler(logging.Handler):  # Inherit from logging.Handler
    ''' Writes out logs to the reporst window '''

    def __init__(self, add_line, show_level=True, short_level=True):
        # run the regular Handler __init__
        logging.Handler.__init__(self)

        self.add_line = add_line  # Function for adding a line to the view
        self.show_level = show_level  # Dynamically show levels
        self.short_level = short_level  # Default to true, changed by properties

        self.formatLevelLength = self.formatLevelShort
        if not short_level:
            self.formatLevelLength = self.formatLevelLong

    def emit(self, record):
        # Just handle all formatting here
        if self.show_level:
            level = self.formatLevel(record.levelname)
            message = html.escape(record.msg)
            output = self.formatOutput()
            self.add_line(output.format(level, message))
        else:
            message = html.escape(record.msg)
            output = self.formatOutput()
            self.add_line(output.format(message))

    def formatLevel(self, levelname):
        output = "{0}{1}{2}"
        level = self.formatLevelLength(levelname)
        if levelname == "INFO":
            return output.format("<font color=\"Green\"><b>", level, "</b></font>")
        elif levelname == "WARNING":
            return output.format("<font color=\"Orange\"><b>", level, "</b></font>")
        elif levelname == "ERROR":
            return output.format("<font color=\"Red\"><b>", level, "</b></font>")
        elif levelname == "CRITICAL":
            return output.format("<font color=\"Red\"><b>", level, "</b></font>")
        else:
            return output.format("<font color=\"Blue\"><b>", level, "</b></font>")

    def formatLevelShort(self, levelname):
        return f'[{levelname[0:1]}]'

    def formatLevelLong(self, levelname):
        output = "{0:<10}"
        if levelname in ["DEBUG", "INFO", "WARNING"]:
            return output.format(f'[{levelname.capitalize()}]')
        else:
            return output.format(f'[{levelname.upper()}]')

    def formatOutput(self):
        ''' Returns the correct output format based on internal settings '''
        if self.show_level:
            if self.short_level:
                return "<tr><td width=\"25\">{0}</td><td><pre>{1}</pre></td></tr>"
            return "<tr><td width=\"75\">{0}</td><td><pre>{1}</pre></td></tr>"
        return "<tr><td><pre>{0}</pre></td></tr>"



"""
Copyright 2008-2015 Free Software Foundation, Inc.
This file is part of GNU Radio

SPDX-License-Identifier: GPL-2.0-or-later

"""

import collections

from .base import Element
from .Constants import ALIASES_OF
from .utils.descriptors import lazy_property


class Connection(Element):
    """
    Stores information about a connection between two block ports. This class
    knows:
    - Where the source and sink ports are (on which blocks)
    - The domain (message, stream, ...)
    - Which parameters are associated with this connection
    """

    is_connection = True
    category = []
    documentation = {'': ''}
    doc_url = ''

    def __init__(self, parent, source, sink):
        """
        Make a new connection given the parent and 2 ports.

        Args:
            parent: the parent of this element (a flow graph)
            source: a port (any direction)
            sink: a port (any direction)
        @throws Error cannot make connection

        Returns:
            a new connection
        """
        Element.__init__(self, parent)

        if not source.is_source:
            source, sink = sink, source
        if not source.is_source:
            raise ValueError('Connection could not isolate source')
        if not sink.is_sink:
            raise ValueError('Connection could not isolate sink')

        self.source_port = source
        self.sink_port = sink

        # Unlike the blocks, connection parameters are defined in the connection
        # domain definition files, as all connections within the same domain
        # share the same properties.
        param_factory = self.parent_platform.make_param
        conn_parameters = self.parent_platform.connection_params.get(self.type, {})
        self.params = collections.OrderedDict(
            (data['id'], param_factory(parent=self, **data))
            for data in conn_parameters
        )

    def __str__(self):
        return 'Connection (\n\t{}\n\t\t{}\n\t{}\n\t\t{}\n)'.format(
            self.source_block, self.source_port, self.sink_block, self.sink_port,
        )

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.source_port == other.source_port and self.sink_port == other.sink_port

    def __hash__(self):
        return hash((self.source_port, self.sink_port))

    def __iter__(self):
        return iter((self.source_port, self.sink_port))

    def children(self):
        """ This includes the connection parameters """
        return self.params.values()

    @lazy_property
    def source_block(self):
        return self.source_port.parent_block

    @lazy_property
    def sink_block(self):
        return self.sink_port.parent_block

    @property
    def type(self):
        return self.source_port.domain, self.sink_port.domain

    @property
    def enabled(self):
        """
        Get the enabled state of this connection.

        Returns:
            true if source and sink blocks are enabled
        """
        return self.source_block.enabled and self.sink_block.enabled

    @property
    def label(self):
        """ Returns a label for dialogs """
        src_domain, sink_domain = [
            self.parent_platform.domains[d].name for d in self.type]
        return f'Connection ({src_domain} → {sink_domain})'

    @property
    def namespace_templates(self):
        """Returns everything we want to have available in the template rendering"""
        return {key: param.template_arg for key, param in self.params.items()}

    def validate(self):
        """
        Validate the connections.
        The ports must match in io size.
        """
        Element.validate(self)
        platform = self.parent_platform

        if self.type not in platform.connection_templates:
            self.add_error_message('No connection known between domains "{}" and "{}"'
                                   ''.format(*self.type))

        source_dtype = self.source_port.dtype
        sink_dtype = self.sink_port.dtype
        if source_dtype != sink_dtype and source_dtype not in ALIASES_OF.get(
            sink_dtype, set()
        ):
            self.add_error_message('Source IO type "{}" does not match sink IO type "{}".'.format(
                source_dtype, sink_dtype))

        source_size = self.source_port.item_size
        sink_size = self.sink_port.item_size
        if source_size != sink_size:
            self.add_error_message(
                'Source IO size "{}" does not match sink IO size "{}".'.format(source_size, sink_size))

    ##############################################
    # Import/Export Methods
    ##############################################
    def export_data(self):
        """
        Export this connection's info.

        Returns:
            A tuple with connection info, and parameters.
        """
        # See if we need to use file format version 2:
        if self.params:
            return {
                'src_blk_id': self.source_block.name,
                'src_port_id': self.source_port.key,
                'snk_blk_id': self.sink_block.name,
                'snk_port_id': self.sink_port.key,
                'params': collections.OrderedDict(sorted(
                    (param_id, param.value)
                    for param_id, param in self.params.items())),
            }

        # If there's no reason to do otherwise, we can export info as
        # FLOW_GRAPH_FILE_FORMAT_VERSION 1 format:
        return [
            self.source_block.name, self.source_port.key,
            self.sink_block.name, self.sink_port.key,
        ]

    def import_data(self, params):
        """
        Import connection parameters.
        """
        for key, value in params.items():
            try:
                self.params[key].set_value(value)
            except KeyError:
                continue



# TODO: This file is a modified copy of the old gui/Platform.py
"""
Copyright 2008,2013 Free Software Foundation, Inc.
This file is part of GNU Radio
SPDX-License-Identifier: GPL-2.0-or-later
"""


from qtpy import QtGui

from ... import Constants


def get_color(rgb):
    color = QtGui.QColor(rgb)
    return color


#################################################################################
# fg colors
#################################################################################
HIGHLIGHT_COLOR = get_color('#00FFFF')
BORDER_COLOR = get_color('#616161')
BORDER_COLOR_DISABLED = get_color('#888888')
FONT_COLOR = get_color('#000000')

# Deprecated blocks
# a light warm yellow
BLOCK_DEPRECATED_BACKGROUND_COLOR = get_color('#FED86B')
# orange
BLOCK_DEPRECATED_BORDER_COLOR = get_color('#FF540B')

# Missing blocks
MISSING_BLOCK_BACKGROUND_COLOR = get_color('#FFF2F2')
MISSING_BLOCK_BORDER_COLOR = get_color('#FF0000')

# Block color constants
BLOCK_ENABLED_COLOR = get_color('#F1ECFF')
BLOCK_DISABLED_COLOR = get_color('#CCCCCC')
BLOCK_BYPASSED_COLOR = get_color('#F4FF81')

CONNECTION_ENABLED_COLOR = get_color('#616161')
CONNECTION_DISABLED_COLOR = get_color('#BBBBBB')
CONNECTION_ERROR_COLOR = get_color('#FF0000')

DARK_FLOWGRAPH_BACKGROUND_COLOR = get_color('#19232D')
LIGHT_FLOWGRAPH_BACKGROUND_COLOR = get_color('#FFFFFF')


#################################################################################
# port colors
#################################################################################
PORT_TYPE_TO_COLOR = {key: get_color(color) for name, key, sizeof, color in Constants.CORE_TYPES}
PORT_TYPE_TO_COLOR.update((key, get_color(color)) for key, (_, color) in Constants.ALIAS_TYPES.items())
